/* acm-s: a simple processor simulator
   Copyright (C) 2003 Angelo "Encelo" Theodorou
*/


#include "acm-s.h"
#include "funcs.h"


/* Ritorna una parola di 16 bit componendo due byte */
unsigned short int make_word(unsigned char hi, unsigned char low)
{
	return (hi*16*16 + low);
}


/* Estrae il byte alto da una parola di 16 bit */
unsigned char get_hi_byte(short int word)
{
	return (word >> 8);
}


/* Estrae il byte basso da una parola di 16 bit */
unsigned char get_low_byte(short int word)
{
	return (word - (word >> 8)*16*16);
}


/* Ritorna 1 se l'istruzione instr ha bisogno di un operando, altrimenti 0 */
int need_operand(unsigned short int instr)
{
	if(instr > 0x1000 && instr < 0x2fff)
		return 1;
	else
		return 0;
}


/* Azzera il contenuto della memoria. Ritorna sempre 0 */
int wipe_mem(void)
{
	extern char mem[];
	int i;                                                                                                                      
	for(i=0; i < MAXMEM; i++)
		mem[i] = 0;
	return 0;
}


/* Ritorna il bit in posizione p di x */
unsigned getbit(unsigned x, int p)
{
	return (x >> p) & ~(~0 << 1);
}


char mnemonic[10];

/* Disassembla un'istruzione ricavando il codice mnemonico. Ritorna sempre 0 */
int disasm(unsigned short int instr)
{
	if(instr == 0x1111)
		strcpy(mnemonic, "LOA A N\0");
	else if(instr == 0x1112)
		strcpy(mnemonic, "LOA A ADR\0");
	else if(instr == 0x1113)
		strcpy(mnemonic, "LOA A I\0");
	else if(instr == 0x1212)
		strcpy(mnemonic, "STO A ADR\0");
	else if(instr == 0x1213)
		strcpy(mnemonic, "STO A I\0");
	else if(instr == 0x1311)
		strcpy(mnemonic, "ADD A N\0");
	else if(instr == 0x1312)
		strcpy(mnemonic, "ADD A ADR\0");
	else if(instr == 0x1313)
		strcpy(mnemonic, "ADD A I\0");
	else if(instr == 0x1411)
		strcpy(mnemonic, "ADC A N\0");
	else if(instr == 0x1412)
		strcpy(mnemonic, "ADC A ADR\0");
	else if(instr == 0x1413)
		strcpy(mnemonic, "ADC A I\0");
	else if(instr == 0x1511)
		strcpy(mnemonic, "SUB A N\0");
	else if(instr == 0x1512)
		strcpy(mnemonic, "SUB A ADR\0");
	else if(instr == 0x1513)
		strcpy(mnemonic, "SUB A I\0");
	else if(instr == 0x1711)
		strcpy(mnemonic, "ADD A N\0");
	else if(instr == 0x1712)
		strcpy(mnemonic, "ADD A ADR\0");
	else if(instr == 0x1713)
		strcpy(mnemonic, "ADD A I\0");
	else if(instr == 0x1811)
		strcpy(mnemonic, "ORR A N\0");
	else if(instr == 0x1812)
		strcpy(mnemonic, "ORR A ADR\0");
	else if(instr == 0x1813)
		strcpy(mnemonic, "ORR A I\0");
	else if(instr == 0x1911)
		strcpy(mnemonic, "NEQ A N\0");
	else if(instr == 0x1912)
		strcpy(mnemonic, "NEQ A ADR\0");
	else if(instr == 0x1913)
		strcpy(mnemonic, "NEQ A I\0");
	else if(instr == 0x1a11)
		strcpy(mnemonic, "CMP A N\0");
	else if(instr == 0x1a12)
		strcpy(mnemonic, "CMP A ADR\0");
	else if(instr == 0x1a13)
		strcpy(mnemonic, "CMP A I\0");
	else if(instr == 0x0110)
		strcpy(mnemonic, "CLR A\0");
	else if(instr == 0x0210)
		strcpy(mnemonic, "INC A\0");
	else if(instr == 0x0310)
		strcpy(mnemonic, "DEC A\0");
	else if(instr == 0x0910)
		strcpy(mnemonic, "NEG A\0");
	else if(instr == 0x5100)
		strcpy(mnemonic, "JMP\0");
	else if(instr >= 0x6100 && instr <= 0x61ff)
		sprintf(mnemonic, "%s 0x%.2x\0", "BRN", (unsigned char)(instr - 0x6100));
	else if(instr >= 0x6200 && instr <= 0x62ff)
		sprintf(mnemonic, "%s 0x%.2x\0", "BZE", (unsigned char)(instr - 0x6200));
	else if(instr >= 0x6300 && instr <= 0x63ff)
		sprintf(mnemonic, "%s 0x%.2x\0", "BNE", (unsigned char)(instr - 0x6300));
	else if(instr == 0x7110)
		strcpy(mnemonic, "IRQ T\0");
	else if(instr == 0x7210)
		strcpy(mnemonic, "INP A\0");
	else if(instr == 0x7310)
		strcpy(mnemonic, "ORQ T\0");
	else if(instr == 0x7410)
		strcpy(mnemonic, "OUP A\0");
	else if(instr == 0x8400)
		strcpy(mnemonic, "NUL\0");
	else if(instr == 0x8500)
		strcpy(mnemonic, "HLT\0");
	else
		strcpy(mnemonic, "UNK\0");

	return 0;
}
