/* acm-s: a simple processor simulator
   Copyright (C) 2003 Angelo "Encelo" Theodorou
*/

#include "acm-s.h"
#include "run.h"
#include "funcs.h"


/* Variabili esterne che implementano i registri dell'acm-s */
extern char mem[];                                   /* Memoria RAM */
unsigned short int pc;                               /* Program Counter */
unsigned short int reg_instr;                        /* Registro di istruzione */
unsigned short int reg_addr_mem;                     /* Registro di indirizzo memoria */
short int reg_data_mem;                              /* Registro dati memoria */
short int accum;                                     /* Accumulatore */
short int zero, carry, negative, overflow;           /* Codici di condizione */
char input, output;                                  /* Registri di input/output */
/* -------------------------------------------------------- */
int reg_dec[4];                                      /* Registro di decodifica istruzione (ausiliario) */
int alt;                                             /* Flag di fine esecuzione (ausiliario) */

/* Azzera i contenuti di tutti i registri. Ritorna sempre 0 */
int reset(void)
{
	int i;

	pc = reg_instr = reg_addr_mem = reg_data_mem = 0;
	accum = 0;
	zero = carry = negative = overflow = 0;
	input = output = 0;
	for(i=0; i <= 3 ; i++)
		reg_dec[i] = 0;
	alt = 0;
	return 0;
}


/* Esegue il programma normalmente o tracciandolo, in dipendenza della trace_cond. */
/* Ritorna 1 se  in "trace", 0 se  in esecuzione normale */
int run(int trace_cond)
{
	while(alt != 1){
		if (pc%2) {
			printf("Instruction at odd address: 0x%.4hx!!!\n", pc);
			exit(-5);
		}
		reg_instr = fetch_word();
		decode();
		if (execute() == -1)
			printf("\nUndefined instruction: 0x%.4hx!\n", reg_instr);
		if (trace_cond == TRACE){
			show_state();
			return 1;
		}
	}
	return 0;
}


/* Incrementa il PC e ritorna una parola di macchina */
short int fetch_word(void)
{
	pc += 2;
	return make_word(mem[pc-2], mem[pc-1]);
}


/* Sistema nell'array reg_dec[] le cifre esadecimali che compongono l'istruzione. Ritorna sempre 0 */
int decode(void)
{
	reg_dec[0] = reg_instr / 0x1000;
	reg_dec[1] = reg_instr % 0x1000 / 0x0100;
	reg_dec[2] = reg_instr % 0x0100 / 0x0010;
	reg_dec[3] = reg_instr % 0x0010 / 0x0001;

	if (need_operand(reg_instr))
		reg_addr_mem = fetch_word(); /* Fetching dell'operando per l'istruzione */
	return 0;
}


/* Controlla che l'indirizzo in input sia valido e non dispari. Ritorna 0 o esce. */
int check_addr(int addr)
{
	if (addr > MAXMEM){
		printf("Out of memory range: 0x%.4hx!!!\n", addr);
		exit(-5);
	}
	else if (addr % 2){
		printf("Accessing odd address: 0x%.4hx!!!\n", addr);
		exit(-5);
	}
	return 0;
}
	

/* Controlla lo stato zero per il registro */
int check_z(short int reg)
{
	if (reg == 0)
		return 1;
	else
		return 0;
}


/* Controlla lo stato negative per il registro */
int check_n(short int reg)
{
	if (reg < 0)
		return 1;
	else
		return 0;
}


/* Controlla il bit di carry nella somma di due registri a 16 bit */
int check_c(short int reg1, short int reg2)
{
	if (getbit(reg1, 15) && getbit(reg2, 15))
		return 1;
	else
		return 0;
}


/* Controlla la condizione di overflow nella somma di due registri a 16 bit */
int check_v(short int reg1, short int reg2)
{
		if ((getbit(reg1, 15) == getbit(reg2, 15)) && (getbit(reg1 + reg2, 15) != getbit(reg1, 15)))
			return 1;
		else
			return 0;			
}


/* Esegue l'istruzione contenuta nel reg_instr avvalendosi dell'array reg_dec[] */
/* Ritorna 0 nel caso di corretta esecuzione o -1 in caso di istruzione sconosciuta */
int execute(void)
{
	short int temp; /* Registro temporaneo di servizio */
	
	switch(reg_dec[0]) {
		case 1: /* GRUPPO DI INDIRIZZAMENTO MEMORIA */
			if (reg_dec[1] != 2) /* Fetch di un dato dalla RAM se non  un istruzione di memorizzazione DA reg */
				switch(reg_dec[3]) { /* Modo di indirizzamento, VERSO registro */
					case 1:
						reg_data_mem = reg_addr_mem;
						break;
					case 2:
						check_addr(reg_addr_mem);
						reg_data_mem = make_word(mem[reg_addr_mem], mem[reg_addr_mem+1]);
						break;
					case 3:
						check_addr(reg_addr_mem);
						check_addr(mem[reg_addr_mem]);
						reg_data_mem = make_word(mem[mem[reg_addr_mem]], mem[mem[reg_addr_mem]+1]);
						break;
					default:
						return -1;
				}
			switch(reg_dec[1]) { /* Azione da svolgere */
				case 1: /* Carica la parola del dato nel registro */
					switch(reg_dec[2]) {
						case 1: /* Accumulatore */
							accum = reg_data_mem;
							zero = check_z(accum);
							negative = check_n(accum);
							break;
						default:
							return -1;
					}
					break;
				case 2: /* Memorizza la parola del registro in memoria */
					switch(reg_dec[2]) {
						case 1: /* Accumulatore */
							switch(reg_dec[3]) { /* Modo di indirizzamento, DA registro */
								case 2:
									zero = check_z(accum);
									negative = check_n(accum);
									check_addr(reg_addr_mem);
									mem[reg_addr_mem] = get_hi_byte(accum);
									mem[reg_addr_mem+1] = get_low_byte(accum);
									break;
								case 3:
									zero = check_z(accum);
									negative = check_n(accum);
									check_addr(mem[reg_addr_mem]);
									check_addr(reg_addr_mem);
									mem[
									make_word(mem[reg_addr_mem], mem[reg_addr_mem+1])
									] = get_hi_byte(accum); 
									mem[
									make_word(mem[reg_addr_mem], mem[reg_addr_mem+1])
									+ 1] = get_low_byte(accum);
									break;
								default:
									return -1;
							}
							break;
						default:
							return -1;
					}
					break;
				case 3: /* Aggiunge la parola del dato al registro */
					switch(reg_dec[2]) {
						case 1: /* Accumulatore */
							carry = check_c(accum, reg_data_mem);
							overflow = check_v(accum, reg_data_mem);
							accum += reg_data_mem;
							zero = check_z(accum);
							negative = check_n(accum);
							break;
						default:
							return -1;
					}
					break;
				case 4: /* Aggiunge la parola del dato ed il bit di riporto al registro */
					switch(reg_dec[2]) {
						case 1: /* Accumulatore */
							carry = check_c(accum, reg_data_mem);
							overflow = check_v(accum, reg_data_mem);
							accum += reg_data_mem + carry;
							zero = check_z(accum);
							negative = check_n(accum);
							break;
						default:
							return -1;
					}
					break;
				case 5: /* Sottrae la parola del dato al registro */
					switch(reg_dec[2]) {
						case 1: /* Accumulatore */
							carry = check_c(accum, -reg_data_mem);
							overflow = check_v(accum, -reg_data_mem);
							accum -= reg_data_mem;
							zero = check_z(accum);
							negative = check_n(accum);
							break;
						default:
							return -1;
					}
					break;
				case 7: /* Fa un AND del dato con il registro */
					switch(reg_dec[2]) {
						case 1: /* Accumulatore */
							accum = accum & reg_data_mem;
							zero = check_z(accum);
							negative = check_n(accum);
							carry = overflow = 0;
							break;
						default:
							return -1;
					}
					break;
				case 8: /* Fa un OR del dato con il registro */
					switch(reg_dec[2]) {
						case 1: /* Accumulatore */
							accum = accum | reg_data_mem;
							zero = check_z(accum);
							negative = check_n(accum);
							carry = overflow = 0;
							break;
						default:
							return -1;
					}
					break;
				case 9: /* Fa un OR esclusivo del dato con il registro */
					switch(reg_dec[2]) {
						case 1: /* Accumulatore */
							accum = accum ^ reg_data_mem;
							zero = check_z(accum);
							negative = check_n(accum);
							carry = overflow = 0;
							break;
						default:
							return -1;
					}
					break;
				case 0xa: /* Confronta il registro con la parola del dato */
					switch(reg_dec[2]) { /* Registro */
						case 1:
							carry = check_c(accum, -reg_data_mem);
							overflow = check_v(accum, -reg_data_mem);
							zero = check_z(accum - reg_data_mem);
							negative = check_n(accum - reg_data_mem);
							break;
						default:
							return -1;
					}
					break;
				default:
					return -1;
			}
			break;
		case 0: /* GRUPPO DI MANIPOLAZIONE DI REGISTRO */
			switch(reg_dec[1]) {
				case 1: /* Azzera registro */
					switch(reg_dec[2]) {
						case 1: /* Accumulatore */
							accum = 0x0000;
							zero = 1;
							negative = carry = overflow = 0;
							break;
						default:
							return -1;
					}
					break;
				case 2: /* Incrementa di 1 il registro */
					switch(reg_dec[2]) {
						case 1: /* Accumulatore */
							carry = check_c(accum, 1);
							overflow = check_v(accum, 1);
							accum++;
							zero = check_z(accum);
							negative = check_n(accum);
							break;
						default:
							return -1;
					}
					break;
				case 3: /* Decrementa di 1 il registro */
					switch(reg_dec[2]) {
						case 1: /* Accumulatore */
							carry = check_c(accum, -1);
							overflow = check_v(accum, -1);
							accum--;
							zero = check_z(accum);
							negative = check_n(accum);
							break;
						default:
							return -1;
					}
					break;
				case 9: /* Nega il registro (operazione NOT) */
					switch(reg_dec[2]) {
						case 1: /* Accumulatore */
							accum = !(accum);
							zero = check_z(accum);
							negative = check_n(accum);
							carry = overflow = 0;
							break;
						default:
							return -1;
					}
					break;
				default:
					return -1;
			}
			break;
		case 5: /* SALTI INCONDIZIONATI */
			switch(reg_dec[1]) {
				case 1:/* Salto incondizionato ad un indirizzo specificato */
					if ((reg_dec[2] != 0) || (reg_dec[3] != 0)) /* se non  proprio 0x5100 */
						return -1;
					pc = fetch_word();
					break;
				default:
					return -1;
			}
			break;
		case 6: /* SALTI DIRAMAZIONALI */
			switch(reg_dec[1]) {
				case 1: /* Salto incondizionato */
					temp = (signed char)(reg_dec[2]*16 + reg_dec[3]);
					pc += temp;
					break;
				case 2: /* Salta se zero */
					if(zero) {
						temp = (signed char)(reg_dec[2]*16 + reg_dec[3]);
						pc += temp;
					}
					break;
				case 3: /* Salta se non zero */
					if(!(zero)) {
						temp = (signed char)(reg_dec[2]*16 + reg_dec[3]);
						pc += temp;
					}
					break;
				default:
					return -1;
			}
			break;
		case 7: /* GRUPPO DI INPUT/OUTPUT */
			switch(reg_dec[1]) {
				case 1: /* Carica dall'unit periferica */
					switch(reg_dec[2]) {
						case 1: /* terminale */
							temp = 0;
							do {
								printf("Insert an hex 8-bit number [0x00-0xff]: ");
								scanf("%hx", &temp);
								if (temp > 0xff)
									printf("Out of bounds!\n");
							} while (temp > 0x00ff);
							input = (signed char)temp;
							break;
						default:
							return -1;
					}
					break;
				case 2: /* Copia il byte dal registro di input */
					switch(reg_dec[2]) {
						case 1: /* accumulatore */
							accum = (short int)input;
							zero = check_z(accum);
							negative = check_n(accum);
							break;
						default:
							return -1;
					}
					break;
				case 3: /* Scarica sull'unita periferica */
					switch(reg_dec[2]) {
						case 1: /* terminale */
							printf("Output hex 8-bit number: 0x%.2hx\n", output);
							break;
						default:
							return -1;
					}
					break;
				case 4: /* Copia il byte nel registro di output */
					switch(reg_dec[2]) {
						case 1: /* accumulatore */
							output = (signed char)accum;
							zero = check_z(accum);
							negative = check_n(accum);
							break;
						default:
							return -1;
					}
					break;
			}
			break;
		case 8: /* OPERAZIONI MISCELLANEE */
			switch(reg_dec[1]) {
				case 4: /* Nessuna operazione */
					if ((reg_dec[2] != 0) || (reg_dec[3] != 0)) /* se non  proprio 0x8400 */
						return -1;
					break;
				case 5: /* Alt */
					if ((reg_dec[2] != 0) || (reg_dec[3] != 0)) /* se non  proprio 0x8500 */
						return -1;
					alt = 1;
					break;	
				default:
					return -1;
			}
			break;
		default:
			return -1;
	}
}


extern char mnemonic[];

/* Mostra lo stato di tutti i registri interni. Ritorna sempre 0 */
int show_state(void)
{
	disasm(reg_instr);
	
	printf("\n\t\t------- Internal State of the acm-s -------\n");
	printf("\t\tpc: 0x%.4hx\n", pc);
	printf("\t\treg_instr: 0x%.4hx     mnemonic: %s\n", reg_instr, mnemonic);
	printf("\t\treg_addr_mem: 0x%.4hx  reg_data_mem: 0x%.4hx\n", reg_addr_mem, reg_data_mem);
	printf("\t\taccum: 0x%.4hx\n", accum);
	printf("\t\tz: %d  n: %d  c: %d  v: %d\n", zero, negative, carry, overflow);
	printf("\t\tinput: 0x%.2x  output: 0x%.2x\n\n", input, output);
	return 0;
}
